/*
 * Ethtool operations for White-Rabbit switch network interface
 *
 * Copyright (C) 2010 CERN (www.cern.ch)
 * Author: Alessandro Rubini <rubini@gnudd.com>
 * Partly from previous work by Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * Partly from previous work by  Emilio G. Cota <cota@braap.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/mii.h>
#include <linux/ethtool.h>
#include <linux/spinlock.h>
#ifdef ETHTOOL_SLINKSETTINGS

/* return false if legacy contained non-0 deprecated fields
 * maxtxpkt/maxrxpkt. rest of ksettings always updated
 */
bool
convert_legacy_settings_to_link_ksettings(
	struct ethtool_link_ksettings *link_ksettings,
	const struct ethtool_cmd *legacy_settings)
{
	bool retval = true;

	memset(link_ksettings, 0, sizeof(*link_ksettings));

	/* This is used to tell users that driver is still using these
	 * deprecated legacy fields, and they should not use
	 * %ETHTOOL_GLINKSETTINGS/%ETHTOOL_SLINKSETTINGS
	 */
	if (legacy_settings->maxtxpkt ||
	    legacy_settings->maxrxpkt)
		retval = false;

	ethtool_convert_legacy_u32_to_link_mode(
		link_ksettings->link_modes.supported,
		legacy_settings->supported);
	ethtool_convert_legacy_u32_to_link_mode(
		link_ksettings->link_modes.advertising,
		legacy_settings->advertising);
	ethtool_convert_legacy_u32_to_link_mode(
		link_ksettings->link_modes.lp_advertising,
		legacy_settings->lp_advertising);
	link_ksettings->base.speed
		= ethtool_cmd_speed(legacy_settings);
	link_ksettings->base.duplex
		= legacy_settings->duplex;
	link_ksettings->base.port
		= legacy_settings->port;
	link_ksettings->base.phy_address
		= legacy_settings->phy_address;
	link_ksettings->base.autoneg
		= legacy_settings->autoneg;
	link_ksettings->base.mdio_support
		= legacy_settings->mdio_support;
	link_ksettings->base.eth_tp_mdix
		= legacy_settings->eth_tp_mdix;
	link_ksettings->base.eth_tp_mdix_ctrl
		= legacy_settings->eth_tp_mdix_ctrl;
	return retval;
}


/* return false if ksettings link modes had higher bits
 * set. legacy_settings always updated (best effort)
 */
static bool
convert_link_ksettings_to_legacy_settings(
	struct ethtool_cmd *legacy_settings,
	const struct ethtool_link_ksettings *link_ksettings)
{
	bool retval = true;

	memset(legacy_settings, 0, sizeof(*legacy_settings));
	/* this also clears the deprecated fields in legacy structure:
	 * __u8		transceiver;
	 * __u32	maxtxpkt;
	 * __u32	maxrxpkt;
	 */

	retval &= ethtool_convert_link_mode_to_legacy_u32(
		&legacy_settings->supported,
		link_ksettings->link_modes.supported);
	retval &= ethtool_convert_link_mode_to_legacy_u32(
		&legacy_settings->advertising,
		link_ksettings->link_modes.advertising);
	retval &= ethtool_convert_link_mode_to_legacy_u32(
		&legacy_settings->lp_advertising,
		link_ksettings->link_modes.lp_advertising);
	ethtool_cmd_speed_set(legacy_settings, link_ksettings->base.speed);
	legacy_settings->duplex
		= link_ksettings->base.duplex;
	legacy_settings->port
		= link_ksettings->base.port;
	legacy_settings->phy_address
		= link_ksettings->base.phy_address;
	legacy_settings->autoneg
		= link_ksettings->base.autoneg;
	legacy_settings->mdio_support
		= link_ksettings->base.mdio_support;
	legacy_settings->eth_tp_mdix
		= link_ksettings->base.eth_tp_mdix;
	legacy_settings->eth_tp_mdix_ctrl
		= link_ksettings->base.eth_tp_mdix_ctrl;
	return retval;
}

static int wrn_get_link_ksettings(struct net_device *dev, struct ethtool_link_ksettings *link_ksettings)
{
	struct wrn_ep *ep = netdev_priv(dev);
	struct ethtool_cmd s_cmd, *cmd;
	int ret=0;

	cmd = &s_cmd;

	spin_lock_irq(&ep->lock);
	mii_ethtool_gset(&ep->mii, cmd);
	spin_unlock_irq(&ep->lock);

	cmd->supported=
		SUPPORTED_FIBRE | /* FIXME: copper sfp? */
		SUPPORTED_Autoneg |
		SUPPORTED_1000baseKX_Full;
	cmd->advertising =
		ADVERTISED_1000baseKX_Full |
		ADVERTISED_Autoneg;
	cmd->port = PORT_FIBRE;
	cmd->speed = SPEED_1000;
	cmd->duplex = DUPLEX_FULL;
	cmd->autoneg = AUTONEG_ENABLE;

	if (!convert_legacy_settings_to_link_ksettings(link_ksettings, cmd))
		return -EINVAL;

	return ret;
}

static int wrn_set_link_ksettings(struct net_device *dev, const struct ethtool_link_ksettings *link_ksettings)
{
	struct wrn_ep *ep = netdev_priv(dev);
	struct ethtool_cmd cmd;
	int ret;

	if (!convert_link_ksettings_to_legacy_settings(&cmd,link_ksettings))
		return -EINVAL;

	spin_lock_irq(&ep->lock);
	ret = mii_ethtool_sset(&ep->mii, &cmd);
	spin_unlock_irq(&ep->lock);

	return ret;
}



#else

static int wrn_get_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
	struct wrn_ep *ep = netdev_priv(dev);
	int ret;

	spin_lock_irq(&ep->lock);
	ret = mii_ethtool_gset(&ep->mii, cmd);
	spin_unlock_irq(&ep->lock);

	cmd->supported=
		SUPPORTED_FIBRE | /* FIXME: copper sfp? */
		SUPPORTED_Autoneg |
		SUPPORTED_1000baseKX_Full;
	cmd->advertising =
		ADVERTISED_1000baseKX_Full |
		ADVERTISED_Autoneg;
	cmd->port = PORT_FIBRE;
	cmd->speed = SPEED_1000;
	cmd->duplex = DUPLEX_FULL;
	cmd->autoneg = AUTONEG_ENABLE;
	return ret;
}

static int wrn_set_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
	struct wrn_ep *ep = netdev_priv(dev);
	int ret;

	spin_lock_irq(&ep->lock);
	ret = mii_ethtool_sset(&ep->mii, cmd);
	spin_unlock_irq(&ep->lock);

	return ret;
}

#endif

static int wrn_nwayreset(struct net_device *dev)
{
	struct wrn_ep *ep = netdev_priv(dev);
	int ret;

	spin_lock_irq(&ep->lock);
	ret = mii_nway_restart(&ep->mii);
	spin_unlock_irq(&ep->lock);

	return ret;
}

static void wrn_get_drvinfo(struct net_device *dev,
			      struct ethtool_drvinfo *info)
{
	strlcpy(info->driver, KBUILD_MODNAME, sizeof(info->driver));
	strlcpy(info->version, DRV_VERSION, sizeof(info->version));
	strlcpy(info->bus_info, dev_name(dev->dev.parent),
		sizeof(info->bus_info));
}

/*
 * These are the operations we support. No coalescing is there since
 * most of the traffic will just happen within the FPGA switching core.
 * Similarly, other funcionality like ringparam are not used.
 * get_eeprom/set_eeprom may be useful for a simple MAC address management.
 */
static const struct ethtool_ops wrn_ethtool_ops = {
#ifdef ETHTOOL_SLINKSETTINGS
	.get_link_ksettings	= wrn_get_link_ksettings,
	.set_link_ksettings	= wrn_set_link_ksettings,
#else
	.get_settings	= wrn_get_settings,
	.set_settings	= wrn_set_settings,
#endif
	.get_drvinfo	= wrn_get_drvinfo,
	.nway_reset	= wrn_nwayreset,
	/* Some of the default methods apply for us */
	.get_link	= ethtool_op_get_link,
	/* FIXME: get_regs_len and get_regs may be useful for debugging */
};

int wrn_ethtool_init(struct net_device *netdev)
{
	netdev->ethtool_ops = &wrn_ethtool_ops;
	return 0;
}
